@@ -9,6 +9,7 @@ from enum import Enum
 from typing import Any, Dict, List, Optional, Tuple, Union, cast
 
 import numpy as np
+import sympy as sp
 from pyspark.sql import DataFrame, SparkSession
 from pyspark.sql import functions as sf
 from pyspark.sql.types import StructType
@@ -27,6 +28,7 @@ from tmlt.core.measurements.composition import Composition
 from tmlt.core.measurements.converters import PureDPToRhoZCDP
 from tmlt.core.measurements.noise_mechanisms import (
     AddDiscreteGaussianNoise,
+    AddGaussianNoise,
     AddGeometricNoise,
     AddLaplaceNoise,
 )
@@ -36,7 +38,11 @@ from tmlt.core.measurements.pandas_measurements.series import (
     NoisyQuantile,
 )
 from tmlt.core.measurements.postprocess import PostProcess
-from tmlt.core.measurements.spark_measurements import AddNoiseToColumn, ApplyInPandas
+from tmlt.core.measurements.spark_measurements import (
+    AddNoiseToColumn,
+    ApplyInPandas,
+    GeometricPartitionSelection,
+)
 from tmlt.core.measures import PureDP, RhoZCDP
 from tmlt.core.metrics import (
     HammingDistance,
@@ -57,6 +63,7 @@ from tmlt.core.transformations.spark_transformations.map import (
     Map,
     RowToRowTransformation,
 )
+from tmlt.core.utils.distributions import double_sided_geometric_inverse_cmf
 from tmlt.core.utils.exact_number import ExactNumber, ExactNumberInput
 from tmlt.core.utils.misc import get_nonconflicting_string
 from tmlt.core.utils.parameters import calculate_noise_scale
@@ -68,6 +75,35 @@ class NoiseMechanism(Enum):
     LAPLACE = 1
     GEOMETRIC = 2
     DISCRETE_GAUSSIAN = 3
+    GAUSSIAN = 4
+
+    def check_output_measure(self, output_measure: Union[PureDP, RhoZCDP]) -> None:
+        """Checks if the specified output measure is supported."""
+        if output_measure not in self.supported_output_measure():
+            str_output_measures = ", ".join(
+                [repr(measure) for measure in self.supported_output_measure()]
+            )
+            raise ValueError(
+                f"Output measure {output_measure}"
+                f" is not supported by noise mechanism {self}."
+                f" Supported output measures are {str_output_measures}."
+            )
+
+    def supported_output_measure(self) -> List[Union[PureDP, RhoZCDP]]:
+        """Returns a list of output measures supported by this noise mechanism."""
+        if self == NoiseMechanism.LAPLACE:
+            return [PureDP(), RhoZCDP()]
+        if self == NoiseMechanism.GEOMETRIC:
+            return [PureDP(), RhoZCDP()]
+        if self == NoiseMechanism.DISCRETE_GAUSSIAN:
+            return [RhoZCDP()]
+        if self == NoiseMechanism.GAUSSIAN:
+            return [RhoZCDP()]
+        raise ValueError(f"Unknown noise mechanism {self}")
+
+    def __str__(self) -> str:
+        """Returns a string representation of the noise mechanism."""
+        return self.name.lower().capitalize().replace("_", " ")
 
 
 @typechecked
@@ -115,13 +151,7 @@ def create_count_measurement(
     """
     d_in = ExactNumber(d_in)
     d_out = ExactNumber(d_out)
-    if (
-        noise_mechanism == NoiseMechanism.DISCRETE_GAUSSIAN
-        and output_measure != RhoZCDP()
-    ):
-        raise ValueError(
-            "Discrete Gaussian noise mechanism can only be used with RhoZCDP."
-        )
+    noise_mechanism.check_output_measure(output_measure)
     count_aggregation: Transformation
     if groupby_transformation is None:
         if isinstance(input_metric, IfGroupedBy):
@@ -145,15 +175,24 @@ def create_count_measurement(
             )
         elif noise_mechanism == NoiseMechanism.GEOMETRIC:
             add_noise_to_number = AddGeometricNoise(alpha=noise_scale)
-        else:
-            assert noise_mechanism == NoiseMechanism.DISCRETE_GAUSSIAN
+        elif noise_mechanism == NoiseMechanism.DISCRETE_GAUSSIAN:
             add_noise_to_number = AddDiscreteGaussianNoise(
                 sigma_squared=noise_scale ** 2
             )
+        elif noise_mechanism == NoiseMechanism.GAUSSIAN:
+            add_noise_to_number = AddGaussianNoise(
+                sigma_squared=noise_scale ** 2, input_domain=NumpyIntegerDomain()
+            )
+        else:
+            raise ValueError(
+                f"Unrecognized noise mechanism {noise_mechanism}. "
+                "Supported noise mechanisms are LAPLACE, "
+                "GEOMETRIC, GAUSSIAN, and DISCRETE_GAUSSIAN."
+            )
         count_measurement: Measurement = count_aggregation | add_noise_to_number
         if (
             output_measure == RhoZCDP()
-            and noise_mechanism != NoiseMechanism.DISCRETE_GAUSSIAN
+            and PureDP() in noise_mechanism.supported_output_measure()
         ):
             # count_measurement has output_measure PureDP and needs to be wrapped in a
             # converter.
@@ -179,11 +218,24 @@ def create_count_measurement(
         )
     elif noise_mechanism == NoiseMechanism.GEOMETRIC:
         add_noise_to_series = AddNoiseToSeries(AddGeometricNoise(alpha=noise_scale))
-    else:
-        assert noise_mechanism == NoiseMechanism.DISCRETE_GAUSSIAN
+    elif noise_mechanism == NoiseMechanism.DISCRETE_GAUSSIAN:
         add_noise_to_series = AddNoiseToSeries(
             AddDiscreteGaussianNoise(sigma_squared=noise_scale ** 2)
         )
+    elif noise_mechanism == NoiseMechanism.GAUSSIAN:
+        add_noise_to_series = AddNoiseToSeries(
+            AddGaussianNoise(
+                sigma_squared=noise_scale ** 2, input_domain=NumpyIntegerDomain()
+            )
+        )
+
+    else:
+        raise ValueError(
+            f"Unrecognized noise mechanism {noise_mechanism}. "
+            "Supported noise mechanisms are LAPLACE, "
+            "GEOMETRIC, GAUSSIAN, and DISCRETE_GAUSSIAN."
+        )
+
     assert isinstance(groupby_count.output_domain, SparkDataFrameDomain)
     add_noise_to_column = AddNoiseToColumn(
         input_domain=groupby_count.output_domain,
@@ -193,7 +245,7 @@ def create_count_measurement(
     count_measurement = groupby_count | add_noise_to_column
     if (
         output_measure == RhoZCDP()
-        and noise_mechanism != NoiseMechanism.DISCRETE_GAUSSIAN
+        and PureDP() in noise_mechanism.supported_output_measure()
     ):
         # count_measurement has output_measure PureDP and needs to be wrapped in a
         # converter.
@@ -250,13 +302,7 @@ def create_count_distinct_measurement(
     """
     d_in = ExactNumber(d_in)
     d_out = ExactNumber(d_out)
-    if (
-        noise_mechanism == NoiseMechanism.DISCRETE_GAUSSIAN
-        and output_measure != RhoZCDP()
-    ):
-        raise ValueError(
-            "Discrete Gaussian noise mechanism can only be used with RhoZCDP."
-        )
+    noise_mechanism.check_output_measure(output_measure)
     count_distinct_aggregation: Transformation
     if groupby_transformation is None:
         if isinstance(input_metric, IfGroupedBy):
@@ -280,22 +326,27 @@ def create_count_distinct_measurement(
             )
         elif noise_mechanism == NoiseMechanism.GEOMETRIC:
             add_noise_to_number = AddGeometricNoise(alpha=noise_scale)
-        else:
-            if noise_mechanism != NoiseMechanism.DISCRETE_GAUSSIAN:
-                raise ValueError(
-                    f"Unrecognized noise mechanism {noise_mechanism}. "
-                    "Supported noise mechanisms are LAPLACE, "
-                    "GEOMETRIC, and DISCRETE_GAUSSIAN."
-                )
+        elif noise_mechanism == NoiseMechanism.DISCRETE_GAUSSIAN:
             add_noise_to_number = AddDiscreteGaussianNoise(
                 sigma_squared=noise_scale ** 2
             )
+        elif noise_mechanism == NoiseMechanism.GAUSSIAN:
+            add_noise_to_number = AddGaussianNoise(
+                sigma_squared=noise_scale ** 2, input_domain=NumpyIntegerDomain()
+            )
+        else:
+            raise ValueError(
+                f"Unrecognized noise mechanism {noise_mechanism}. "
+                "Supported noise mechanisms are LAPLACE, "
+                "GEOMETRIC, GAUSSIAN, and DISCRETE_GAUSSIAN."
+            )
+
         count_distinct_measurement: Measurement = (
             count_distinct_aggregation | add_noise_to_number
         )
         if (
             output_measure == RhoZCDP()
-            and noise_mechanism != NoiseMechanism.DISCRETE_GAUSSIAN
+            and PureDP() in noise_mechanism.supported_output_measure()
         ):
             # the measurement created above has output_measure PureDP,
             # so it needs to be converted
@@ -331,16 +382,22 @@ def create_count_distinct_measurement(
         )
     elif noise_mechanism == NoiseMechanism.GEOMETRIC:
         add_noise_to_series = AddNoiseToSeries(AddGeometricNoise(alpha=noise_scale))
-    else:
-        if noise_mechanism != NoiseMechanism.DISCRETE_GAUSSIAN:
-            raise ValueError(
-                f"Unrecognized noise mechanism {noise_mechanism}. "
-                "Supported noise mechanisms are LAPLACE, "
-                "GEOMETRIC, and DISCRETE_GAUSSIAN."
-            )
+    elif noise_mechanism == NoiseMechanism.DISCRETE_GAUSSIAN:
         add_noise_to_series = AddNoiseToSeries(
             AddDiscreteGaussianNoise(sigma_squared=noise_scale ** 2)
         )
+    elif noise_mechanism == NoiseMechanism.GAUSSIAN:
+        add_noise_to_series = AddNoiseToSeries(
+            AddGaussianNoise(
+                sigma_squared=noise_scale ** 2, input_domain=NumpyIntegerDomain()
+            )
+        )
+    else:
+        raise ValueError(
+            f"Unrecognized noise mechanism {noise_mechanism}. "
+            "Supported noise mechanisms are LAPLACE, "
+            "GEOMETRIC, GAUSSIAN, and DISCRETE_GAUSSIAN."
+        )
     assert isinstance(groupby_count_distinct.output_domain, SparkDataFrameDomain)
     add_noise_to_column = AddNoiseToColumn(
         input_domain=groupby_count_distinct.output_domain,
@@ -350,7 +407,7 @@ def create_count_distinct_measurement(
     count_distinct_measurement = groupby_count_distinct | add_noise_to_column
     if (
         output_measure == RhoZCDP()
-        and noise_mechanism != NoiseMechanism.DISCRETE_GAUSSIAN
+        and PureDP() in noise_mechanism.supported_output_measure()
     ):
         # the count_distinct_measurement generated above has the
         # output_measure PureDP, and needs to be converted
@@ -412,13 +469,7 @@ def create_sum_measurement(
     upper = ExactNumber(upper)
     d_in = ExactNumber(d_in)
     d_out = ExactNumber(d_out)
-    if (
-        noise_mechanism == NoiseMechanism.DISCRETE_GAUSSIAN
-        and output_measure != RhoZCDP()
-    ):
-        raise ValueError(
-            "Discrete Gaussian noise mechanism can only be used with RhoZCDP."
-        )
+    noise_mechanism.check_output_measure(output_measure)
     sum_aggregation: Transformation
     measure_column_domain = input_domain[measure_column].to_numpy_domain()
     if not isinstance(measure_column_domain, (NumpyIntegerDomain, NumpyFloatDomain)):
@@ -448,15 +499,25 @@ def create_sum_measurement(
             )
         elif noise_mechanism == NoiseMechanism.GEOMETRIC:
             add_noise_to_number = AddGeometricNoise(alpha=noise_scale)
-        else:
-            assert noise_mechanism == NoiseMechanism.DISCRETE_GAUSSIAN
+
+        elif noise_mechanism == NoiseMechanism.DISCRETE_GAUSSIAN:
             add_noise_to_number = AddDiscreteGaussianNoise(
                 sigma_squared=noise_scale ** 2
             )
+        elif noise_mechanism == NoiseMechanism.GAUSSIAN:
+            add_noise_to_number = AddGaussianNoise(
+                sigma_squared=noise_scale ** 2, input_domain=measure_column_domain
+            )
+        else:
+            raise ValueError(
+                f"Unrecognized noise mechanism d P {noise_mechanism}. "
+                "Supported noise mechanisms are LAPLACE, "
+                "GEOMETRIC, GAUSSIAN, and DISCRETE_GAUSSIAN."
+            )
         sum_measurement: Measurement = sum_aggregation | add_noise_to_number
         if (
             output_measure == RhoZCDP()
-            and noise_mechanism != NoiseMechanism.DISCRETE_GAUSSIAN
+            and PureDP() in noise_mechanism.supported_output_measure()
         ):
             # sum_measurement has output_measure PureDP and needs to be wrapped in a
             # converter.
@@ -485,11 +546,22 @@ def create_sum_measurement(
         )
     elif noise_mechanism == NoiseMechanism.GEOMETRIC:
         add_noise_to_series = AddNoiseToSeries(AddGeometricNoise(alpha=noise_scale))
-    else:
-        assert noise_mechanism == NoiseMechanism.DISCRETE_GAUSSIAN
+    elif noise_mechanism == NoiseMechanism.DISCRETE_GAUSSIAN:
         add_noise_to_series = AddNoiseToSeries(
             AddDiscreteGaussianNoise(sigma_squared=noise_scale ** 2)
         )
+    elif noise_mechanism == NoiseMechanism.GAUSSIAN:
+        add_noise_to_series = AddNoiseToSeries(
+            AddGaussianNoise(
+                sigma_squared=noise_scale ** 2, input_domain=measure_column_domain
+            )
+        )
+    else:
+        raise ValueError(
+            f"Unrecognized noise mechanism {noise_mechanism}. "
+            "Supported noise mechanisms are LAPLACE, "
+            "GEOMETRIC, GAUSSIAN, and DISCRETE_GAUSSIAN."
+        )
     assert isinstance(sum_aggregation.output_domain, SparkDataFrameDomain)
     add_noise_to_column = AddNoiseToColumn(
         input_domain=sum_aggregation.output_domain,
@@ -499,7 +571,7 @@ def create_sum_measurement(
     sum_measurement = groupby_sum | add_noise_to_column
     if (
         output_measure == RhoZCDP()
-        and noise_mechanism != NoiseMechanism.DISCRETE_GAUSSIAN
+        and PureDP() in noise_mechanism.supported_output_measure()
     ):
         # sum_measurement has output_measure PureDP and needs to be wrapped in a
         # converter.
@@ -1390,3 +1462,76 @@ def _create_map_to_compute_deviations(
         deviations_column,
         squared_deviations_column,
     )
+
+
+@typechecked
+def create_partition_selection_measurement(
+    input_domain: SparkDataFrameDomain,
+    epsilon: ExactNumberInput,
+    delta: ExactNumberInput,
+    d_in: ExactNumberInput = 1,
+    count_column: Optional[str] = None,
+) -> GeometricPartitionSelection:
+    """Returns a partition selection measurement.
+
+    A partition selection measurement created by this function will have a
+    privacy guarantee such that
+    ``measurement.privacy_function(d_in) = (epsilon, delta)``.
+
+    Args:
+        input_domain: Domain of the input Spark DataFrames. Input cannot contain
+            floating point columns.
+        epsilon: The epsilon portion of the (epsilon, delta) privacy budget that
+            you want this measurement to satisfy.
+        delta: The delta portion of the (epsilon, delta) privacy budget that
+            you want this measurement to satisfy.
+        d_in: The given d_in such that
+            ``measurement.privacy_function(d_in) = (epsilon, delta)``.
+        count_column: Column name for output group counts. If None, output column
+            will be named "count".
+    """
+    d_in = ExactNumber(d_in)
+    epsilon = ExactNumber(epsilon)
+    delta = ExactNumber(delta)
+    alpha: ExactNumberInput
+    threshold: int
+
+    if d_in < 1:
+        raise NotImplementedError(
+            "Creating a partition selection measurement with d_in < 1 is not yet"
+            " supported."
+        )
+    if delta > 1 or delta < 0:
+        raise RuntimeError(
+            f"Delta should be between 0 and 1 (inclusive). Instead, it is {delta}"
+        )
+
+    if d_in != 1:
+        # Convert everything to its d_in=1 equivalent
+        epsilon = epsilon / d_in
+        delta = delta / (d_in * sp.E ** (d_in * epsilon))
+
+    # Special case for infinite privacy budget
+    if epsilon == float("inf"):
+        alpha = 0
+        threshold = 9223372036854775807  # largest Long that Spark supports
+    # Normal cases
+    else:
+        alpha = 1 / epsilon
+        threshold = (
+            int(
+                double_sided_geometric_inverse_cmf(
+                    # These both need to be round_up=False
+                    # otherwise you will not get the result you want
+                    (1 - delta).to_float(round_up=False),
+                    alpha.to_float(round_up=False),
+                )
+            )
+            + 2
+        )
+    return GeometricPartitionSelection(
+        input_domain=input_domain,
+        threshold=threshold,
+        alpha=alpha,
+        count_column=count_column,
+    )
